#ifndef BROWNIAN2D_H
#define BROWNIAN2D_H
#include <iostream>
#include <vector>                  // for vector<>
#include <cmath>                   // for exp(), sqrt()
#include <iomanip>                 // for setw(), setfill()
#include <cstdlib>                 // for exit(), system(), srand(), rand()
#include <fstream>
#include <strstream>
#include <ctime>                   // for time(NULL)
#include <cstdio>                  // for sprintf()
#include <string>
#include <map>                     // associate ij object with entries
#include "bubble.h"
#include "lattice_point.h"
#include "ij.h"                    // an object that holds indices of a table.
#include "ran3_long.h"             
#include "ran3_float.h"
#include "mat.h"
#include "engine.h"                // Matlab engine for visualization
#define  MOD(X, Y)     (X & (Y-1))
#define  DS            256         // default square lattice dimension
#define  NB            4           // # of droplets to be created
#define  NRUN          1           // # of simulation runs
#define  NMCS          100         // # of mcs each run
#define  NEminus1      0.01        // data collected every 0.01 mcs
#define  NE0           0.1         // data collected every 0.1 mcs
#define  NE1           1           // data collected every 1 mcs
#define  NE2           10          // data collected every 10 mcs
#define  NE3           100         // data collected every 100 mcs
#define  NE4           1000        // data collected every 1000 mcs
#define  NTsmall       19          // number of times data is collected t<1
#define  NT            56          // total number of times data is collected
#define  NN1           4           // up to 1st nearest neighbours
#define  NN2           8           // up to 2nd nearest neighbours
#define  NN3           12          // up to 3rd nearest neighbours
#define  NN4           20          // up to 4th nearest neighbours
#define  NNn           NN4
using namespace std;

/************************************************************************/
/*         Started on 27/2/2002                                         */
/*         Helical boundary condition                                   */
/*         Dry foam                                                     */
/*         Default size is 128*128;                                     */
/*         Use only powers of 2 for lattice dimensions                  */
/*         Coupling between bubble and medium is 0                      */
/*         Note: in order for MOD() to work, XY must be a power of 2.   */
/*                                                                      */
/*         This is version 2 (21/3/2003)                                */
/*         It includes codes to calculate displacement of droplets      */
/*         as well as the energy and area of droplets.                  */
/*                                                                      */
/*         This is version 3 (14/1/2004)                                */
/*         Incorporates algorithm 2 and 3 by way of virtual function.   */
/*         Default size is DS x DS.                                     */
/*         Uses the result from previous Mathematica program            */
/*         to produce an initial circular droplet.                      */
/*                                                                      */
/*         (8/9/2004)                                                   */
/*         Gather data at smaller t (0.1 mcs, 0.01 mcs)                 */
/*                                                                      */
/*         (12/10/2004)                                                 */
/*         Count the number of spin-flips in each Monte-Carlo steps.    */
/************************************************************************/

class brownian2d {
 protected:
  int X, Y, XY;
  int disp[NN4+1];
  float SpinFlipsCount[12][NMCS+1];
  float countspinflips;
  double kT;
  float Ja;                   // area constraint
  float Jp;                   // perimeter constraint
  float Jij[2][2];            // coupling constants
  float RR[NT][NRUN*NB];
  float meanRR[NT][3];
  long ran_seed1, ran_seed2, ran_seed3;
  std::vector<simple_lattice_point> element; 
  bubble* bub_array[NB+1];    // An array of pointers to droplets
  map<ij, int> ae_count;      // tabulates the area-energy distribution

  void initialize();
  void set_disp();
  void read_variables();
  void reset_spin_count();
  void generate_random_seeds();
  void create_droplets(int);
  void create_droplet_at(int, int, int, int);
  void show_all_lattice_points() const;
  void save_img(const char*, int) const;
  void save_data(const char*, int) const;
  void save_spin_counts(const char*, int) const;
  void commence_droplet_tracking();
  void reset();
  void gather_data_smallt(int&, float, int);
  void gather_data(int&, int, int, Engine* ep1);
  void calc_meanRR();
  virtual bool flip_spin(const float);
  bool flip_routine(simple_lattice_point&, simple_lattice_point&, int&, int);

 public:
  brownian2d(int x = DS) : X(x), Y(x), XY(x*x) { initialize(); }
  brownian2d(int x, int y) : X(x), Y(y), XY(x*y) { initialize(); }
  ~brownian2d() { for(int i=0; i<NB; i++)  delete bub_array[i]; };

  int get_X() const { return X; }
  int get_Y() const { return Y; }
  int get_XY() const { return XY; }
  void show_spins_of_all_lattice_points() const;
  void show_droplets_info() const;
  void minimize(const char*);
};

// derived class (algorithm 2a is in brownian2d class).
// everything in algorithm3 is the same as brownian2d except for the
// flip_spin() function.
class algorithm3 : public brownian2d { 
 protected:
  virtual bool flip_spin(const float);
 public:
  algorithm3(int x = DS) : brownian2d(x) { initialize(); }
  algorithm3(int x, int y) : brownian2d(x, y) { initialize(); }
  ~algorithm3() { for(int i=0; i<NB; i++)  delete bub_array[i]; };
};

class algorithm2b : public brownian2d { 
 protected:
  virtual bool flip_spin(const float);
 public:
  algorithm2b(int x = DS) : brownian2d(x) { initialize(); }
  algorithm2b(int x, int y) : brownian2d(x, y) { initialize(); }
  ~algorithm2b() { for(int i=0; i<NB; i++)  delete bub_array[i]; };
};

void brownian2d::initialize() {
  set_disp();
  read_variables();
  element.reserve(XY);
  generate_random_seeds();
  
  bub_array[0] = new bubble(0, 0);           // spin 0, type 0 (medium)
  bub_array[0]->set_area(0);
  for(int i=1; i<=NB; i++)
    bub_array[i] = new bubble(i, 1);         // spin i, type 1 (droplet)
}

void brownian2d::set_disp() {

  disp[0] = 0;

  // 1st nearest               // 2nd nearest
  disp[1] = 1;                 disp[5] = 1 + X;
  disp[2] = -1;                disp[6] = 1 - X;
  disp[3] = X;                 disp[7] = -1 + X;
  disp[4] = -X;                disp[8] = -1 - X;

  // 3rd nearest               // 4th nearest 
  disp[9] = 2;                 disp[13] = 2 + X;
  disp[10] = -2;               disp[14] = 2 - X;
  disp[11] = 2*X;              disp[15] = 1 + 2*X;
  disp[12] = -2*X;             disp[16] = 1 - 2*X;
                               disp[17] = -2 + X;
                               disp[18] = -2 - X;
                               disp[19] = -1 + 2*X;
                               disp[20] = -1 - 2*X;
}

void brownian2d::read_variables() {

  char comment[100];

  fstream datafile("variables.txt", ios::in);
  if(!datafile) {
    cerr << "Fail to open variables.txt file.\n";
    exit(EXIT_FAILURE);
  }

  datafile >> Ja;          datafile.getline(comment, 100);  // use '\n'
  datafile >> Jp;          datafile.getline(comment, 100);  // as delimiter
  datafile >> Jij[0][0];   datafile.getline(comment, 100);
  datafile >> Jij[0][1];   datafile.getline(comment, 100);
  datafile >> Jij[1][0];   datafile.getline(comment, 100);
  datafile >> Jij[1][1];   datafile.getline(comment, 100);

  datafile.close();
}

void brownian2d::reset_spin_count() {

  for(int i=0; i<12; i++)
    for(int t=0; t<=NMCS; t++)
      SpinFlipsCount[i][t] = 0;
}

void brownian2d::generate_random_seeds() {

  srand(time(NULL));

  ran_seed1 = -rand();
  ran_seed2 = -rand();
  ran_seed3 = -rand();
}

void brownian2d::create_droplets(int radius) {

  // location of droplets
  int xc[NB + 1] = {X/2, X/4, X/4, 3*X/4, 3*X/4};
  int yc[NB + 1] = {Y/2, Y/4, 3*Y/4, Y/4, 3*Y/4}; 

  // declare the droplets including the medium
  for(int i=0; i<=NB; i++) {
    bub_array[i]->set_area(0);
    bub_array[i]->set_sum_xy(0, 0);
    bub_array[i]->set_energy(0);
  }
  // initialize all lattice points to point to medium
  for(int coord=0; coord<XY; coord++) {
    element[coord].set_spin(bub_array[0]);
    element[coord].increase_bubble_area();
  } 
  // create droplet i
  for(int i=1; i<=NB; i++)
    create_droplet_at(radius, i, xc[i], yc[i]);

  // calculate initial droplet energy
  for(int coord=0; coord<XY; coord++)
    for(int n=1; n<=NNn; n++)
      if(element[coord] != element[MOD(coord + disp[n], XY)])
	element[coord].increment_bubble_energy(1);

}

void brownian2d::create_droplet_at(int radius, int i, int xc, int yc) {

  char* filename = new char[80];
  sprintf(filename, "./Circle_file/circle%d.dat", radius);
  fstream circlefile(filename, ios::in);
  if(!circlefile) {
    cerr << "Fail to open file " << filename << endl;
    exit(EXIT_FAILURE);
  }
  // Create the droplet at (xc, yc)
  int pix;
  int x0, y0;
  int x1, y1;
  int target_area = 0;
  for(int x=0; x<radius; x++) {
    circlefile >> pix;
    for(int y=0; y<pix; y++) {
      x0 = x + xc; x1 = -x -1 + xc;
      y0 = y + yc; y1 = -y -1 + yc;
      element[y0*X + x0].set_spin(bub_array[i]);
      element[y0*X + x1].set_spin(bub_array[i]);
      element[y1*X + x0].set_spin(bub_array[i]);
      element[y1*X + x1].set_spin(bub_array[i]);
      target_area += 4;
      bub_array[i]->increase_area(4);
      bub_array[0]->decrease_area(4);
      bub_array[i]->update_sum_xy(x0,y0);
      bub_array[i]->update_sum_xy(x0,y1);
      bub_array[i]->update_sum_xy(x1,y0);
      bub_array[i]->update_sum_xy(x1,y1);
    }
  }
  bub_array[0]->set_target_area(XY - target_area);
  bub_array[i]->set_target_area(target_area);
  circlefile.close();

}

void brownian2d::show_all_lattice_points() const {
  
  for(int j=0; j<Y; j++) {
    cout << endl;
    for(int i=0; i<X; i++)
      cout << element[j*X + i] << " ";
    cout << endl;
  }
  
}

void brownian2d::show_droplets_info() const {
  for(int i=1; i<=NB; i++) {
    cout << "spin = " << bub_array[i]->get_spin() << "\t";
    cout << "type = " << bub_array[i]->get_type() << "\t";
    cout << "target area = " << setw(4) 
	 << bub_array[i]->get_target_area() << "\t";
    cout << "actual area = " << bub_array[i]->get_area() << "\t";
    cout << "energy = " << bub_array[i]->get_energy() << endl;
  }
}

void brownian2d::save_img(const char* argv, int mcs) const {
  char* name = new char[32];

  ostrstream filename(name, 32);
  if(!filename) {
    cerr << "Fail to create filename " << argv << endl;
    exit(EXIT_FAILURE);
  }
  filename.seekp(0);
  filename << "./movies/" << argv;
  filename << ".dat" << ends;
  fstream frame_file(name, ios::out);
  if(!frame_file) {
    cerr << "Fail to open file " << name << endl;
    exit(EXIT_FAILURE);
  }

  for(int i=0; i<XY; i++)  
    frame_file << element[i];
  
  frame_file.close();
  delete name;
}

void brownian2d::save_data(const char* algo, int radius) const{

  char* name = new char[32];

  ostrstream filename(name, 32);
  if(!filename) {
    cerr << "Fail to create filename " << endl;
    exit(EXIT_FAILURE);
  }
  filename.seekp(0);   // diffusion data
  filename << "./data/difusn" << algo << "_" << radius << "_";
  filename << setfill('0') << setw(2) << 10*kT << ".dat" << ends;
  fstream ctr_m_file(name, ios::out);
  if(!ctr_m_file) {
    cerr << "Fail to open file " << name << endl;
    exit(EXIT_FAILURE);
  }
  filename.seekp(0);   // area-energy distribution
  filename << "./data/ae_dist" << algo << "_" << radius << "_";
  filename << setfill('0') << setw(2) << 10*kT << ".dat" << ends;
  fstream ae_file(name, ios::out);
  if(!ae_file) {
    cerr << "Fail to open file " << name << endl;
    exit(EXIT_FAILURE);
  }
  
  filename.seekp(0);   // spin-flips count
  filename << "./data/sp_count" << algo << "_" << radius << ".dat" << ends;
  fstream sp_file(name, ios::out);
  if(!sp_file) {
    cerr << "Fail to open file " << name << endl;
    exit(EXIT_FAILURE);
  }
  for(int t=0; t<NTsmall; t++)
    ctr_m_file << meanRR[t][0] << "\t"
	       << setw(10) << meanRR[t][1] << " "
	       << setw(10) << meanRR[t][2] << endl;
  
  // Normal iterator can be used to modify the contents of maps.
  // Since save_data() is a constant function (doesn't allow 
  // modification of data), the compiler will complain if
  // we use normal iterator. Solution: use const_iterator.
  
  map<ij, int>::const_iterator it;
  for(it=ae_count.begin(); it!=ae_count.end(); it++)
    ae_file << (*it).first.getI() << "\t" << (*it).first.getJ() << "\t"
	    << (*it).second << endl;
  
  ctr_m_file.close();
  ae_file.close();

  delete name;
}

void brownian2d::save_spin_counts(const char* algo, int radius) const {
  char* name = new char[32];

  ostrstream filename(name, 32);
  if(!filename) {
    cerr << "Fail to create filename " << endl;
    exit(EXIT_FAILURE);
  }
  filename.seekp(0);   // spin-flips count
  filename << "./data/sp_count" << algo << "_" << radius << ".dat" << ends;
  fstream sp_file(name, ios::out);
  if(!sp_file) {
    cerr << "Fail to open file " << name << endl;
    exit(EXIT_FAILURE);
  }
  for(int i=0; i<12; i++) {
    for(int t=0; t<=NMCS; t++)
      sp_file << SpinFlipsCount[i][t] << " ";
    sp_file << endl;
  }
  delete name;
}

void brownian2d::commence_droplet_tracking() {

  for(int i=0; i<=NB; i++)
    bub_array[i]->set_initial_ctr();

}

void brownian2d::reset() {

  for(int t=0; t<NT; t++) {
    meanRR[t][0] = meanRR[t][1] = meanRR[t][2] = 0;
    for(int r=0; r<NRUN*NB; r++)
      RR[t][r] = 0;
  }
}

void brownian2d::minimize(const char* algo) {

  string data_directory = "mkdir data";
  string movie_directory = "mkdir movies";
  system(data_directory.c_str());
  system(movie_directory.c_str());

  /*Engine *ep1;
  if(!(ep1 = engOpen("\0"))) {
    cerr << "Can't start MATLAB engine1\n";
    exit(EXIT_FAILURE);
  }
  */
  int rad[8] = {3,4,5,6,7,8,12,16};
  int radius;
  for(int r=0; r<8; r++) {
    radius = rad[r];
    cout << "radius = " << radius << endl;

    reset_spin_count();
    for(kT=0.5; kT<=6.0; kT+=0.5) {
      cout << "\t temp = " << kT << endl;
      reset();
      SpinFlipsCount[int(kT/0.5-1)][0] = kT;

      for(int run=1; run<=NRUN; run++) {
	generate_random_seeds();
	create_droplets(radius);

	for(int wm=1; wm<=200; wm++) // warm-up
	  flip_spin(1.0);

	commence_droplet_tracking();
	int t = 0;
	bool flag = false;

	/*for(float mcs=NEminus1; mcs<=NE0; mcs+=NEminus1){
	  flip_spin(NEminus1);
	  gather_data_smallt(t, mcs, run);
	}
	for(float mcs=NE0+NE0; mcs<NE1; mcs+=NE0){
	  flip_spin(NE0);
	  gather_data_smallt(t, mcs, run);
	}
	*/
	for(int mcs=NE1; mcs<=NMCS; mcs+=NE1) {
	  countspinflips = 0;
	  if(flip_spin(NE1)) {
	    flag = true;    // droplet vanished or touching the edge of lattice
	    cout << "run " << run << "\tmcs " <<  mcs << endl;
	    break;
	    }
	  //gather_data(t, mcs, run, ep1);
	  SpinFlipsCount[int(kT/0.5-1)][mcs] = countspinflips/NB;
	}
	
	if(flag) {          // Repeat this run if flag is true
	  generate_random_seeds();
	  run--;
	  continue;
	}
      }
      //calc_meanRR();
      //save_data(algo, radius);
      //ae_count.clear();     // clear the area-energy distribution table
    }
    save_spin_counts(algo, radius);
  }
  //engClose(ep1);
}

void brownian2d::gather_data_smallt(int& t, float mcs, int run) {

  meanRR[t][0] = mcs;
  for(int i=1; i<=NB; i++) {
    meanRR[t][1] += bub_array[i]->get_rsquare();
    RR[t][NB*(run-1) + i - 1] = bub_array[i]->get_rsquare();
  }
  t++;
}

void brownian2d::gather_data(int& t, int mcs, int run, Engine* ep1) {

  char* action = new char[80];
  // gather area-energy distribution data (allow 100 mcs warmup)
  if(mcs >= 100) {
    ij area_energy;
    for(int i=1; i<=NB; i++) {
      area_energy.setIJ(bub_array[i]->get_area(), bub_array[i]->get_energy());
      ae_count[area_energy]++;
    }
  }
  
  // gather diffusion data
  if((mcs <= 10    && mcs%NE1 == 0) ||
     (mcs <= 100   && mcs%NE2 == 0) ||
     (mcs <= 1000  && mcs%NE3 == 0) ||
     (mcs <= 10000 && mcs%NE4 == 0)) {
    meanRR[t][0] = mcs;
    for(int i=1; i<=NB; i++) {
      meanRR[t][1] += bub_array[i]->get_rsquare();
      RR[t][NB*(run-1) + i - 1] = bub_array[i]->get_rsquare();
      }
    cout << t << "\t" << mcs << endl;
    t++;
    //show_droplets_info();
    /*save_img("frame", mcs);
      sprintf(action, "visualize1 ./movies/frame.dat %d %d %d", mcs, X, Y);
      engEvalString(ep1, action);
    */
  }
  delete action;
}

void brownian2d::calc_meanRR() {

  double errRR;
  for(int t=0; t<NT; t++) {
    meanRR[t][1] /= (NRUN*NB);
    errRR = 0;

    for(int r=0; r<NRUN*NB; r++)
      errRR += pow(RR[t][r] - meanRR[t][1], 2);

    meanRR[t][2] = sqrt(errRR/(NRUN*NB*(NRUN*NB - 1)));
  }
}

bool brownian2d::flip_spin(const float ne) {

  int coord1, coord2, coordj;
  int type1, type2, typej;
  int index2;
  int dN;
  double J_ij = Jij[1][1];
  double fac[NNn];
  double dEJ, dEA;

  for(int dn=0; dn<NNn; dn++)
    fac[dn] = (double(NNn + dn)/double(NNn - dn));

  for(int i=0; i<int(ne*XY); i++) {
    dEJ = dEA = 0;
    dN = NNn;
    coord1 = MOD(ran3l(&ran_seed1), XY);
    index2 = ran3l(&ran_seed2)%NN2 + 1;
    coord2 = MOD(coord1 + disp[index2], XY);

    if(element[coord1] != element[coord2]) {
      type1 = element[coord1].get_type();
      type2 = element[coord2].get_type();

      if(type1 == type2) {
	cout << "droplets are touching each other!\t";
	return true;
      }
      // delta_E due to interaction (spin 1 -> spin 2)
      for(int i=1; i<=NNn; i++) {
	coordj = MOD(coord1 + disp[i], XY);
	if(element[coordj] != element[coord1])
	  dN -= 2;
      }
      dEJ = dN*J_ij;

      // delta_E due to area constraint.
      // Bubble1 - 1 pixel, bubble2 + 1 pixel.
      // There is no area constraint on medium.
      // This is ensured by the value of "type".
      dEA = (Ja*type1*(2*(element[coord1].get_target_area() -
			  element[coord1].get_area()) + 1) +
	     Ja*type2*(-2*(element[coord2].get_target_area() -
			   element[coord2].get_area()) + 1));
      
      // flip spins
      if(dEJ<=0 && dEA<=0){
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if((dEJ<=0 && dEA>0) && ran3f(&ran_seed3) < exp(-dEA/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if((dEJ>0 && dEA<=0) && ran3f(&ran_seed3) < fac[dN]*exp(-dEJ/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if(ran3f(&ran_seed3) < fac[dN]*exp(-(dEJ + dEA)/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
    }
  }
  return false;
}

bool brownian2d::flip_routine(simple_lattice_point& spin1,
			 simple_lattice_point& spin2, int& coord1, int dEJ) {
  spin1.decrease_bubble_area();
  spin2.increase_bubble_area();
  spin1.increment_bubble_energy(dEJ);
  spin2.increment_bubble_energy(dEJ);
  if(spin1.get_area() == 0) {
    cout << "droplet vanished!\t";
    return true;
  }
  int x = coord1%X;
  int y = coord1/X;
  if(x == 0 || x == X-1 || y == 0 || y == Y-1) {
    cout << "droplet at lattice edge!\t";
    return true;
  }
  spin1.update_sum_xy(-x, -y);
  spin2.update_sum_xy(x, y);
  spin1 = spin2;
  countspinflips++;
  return false;      

}

bool algorithm3::flip_spin(const float ne) {

  int coord1, coord2, coordj;
  int type1, type2, typej;
  int index2;
  int dN;
  double J_ij = Jij[1][1];
  double dE, dEJ;

  for(int i=0; i<int(ne*XY); i++) {
    dE = 0;
    dN = NNn;
    coord1 = MOD(ran3l(&ran_seed1), XY);
    index2 = ran3l(&ran_seed2)%NN2 + 1;
    coord2 = MOD(coord1 + disp[index2], XY);

    if(element[coord1] != element[coord2]) {
      type1 = element[coord1].get_type();
      type2 = element[coord2].get_type();

      if(type1 == type2) {
	cout << "droplets are touching each other!\t";
	return true;
      }
      // delta_E due to interaction
      for(int i=1; i<=NNn; i++) {
	coordj = MOD(coord1 + disp[i], XY);
	if(element[coordj] != element[coord1])
	  dN -= 2;
      }
      dE = dN*J_ij;
      dEJ = dE;

      // delta_E due to area constraint.
      // droplet1 - 1 pixel, droplet2 + 1 pixel.
      // There is no area constraint on medium.
      // This is ensured by the value of "type".
      dE += (Ja*type1*(2*(element[coord1].get_target_area() -
			  element[coord1].get_area()) + 1) +
	     Ja*type2*(-2*(element[coord2].get_target_area() -
			   element[coord2].get_area()) + 1));

      // flip spins
      if(dE <= 0) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if(kT > 0 && ran3f(&ran_seed3) < exp(-dE/kT))
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
    }
  }
  return false;
}

bool algorithm2b::flip_spin(const float ne) {

  int coord1, coord2, coordj;
  int type1, type2, typej;
  int index2;
  int dN;
  double J_ij = Jij[1][1];
  double dEJ, dEA;

  for(int i=0; i<int(ne*XY); i++) {
    dEJ = dEA = 0;
    dN = NNn;
    coord1 = MOD(ran3l(&ran_seed1), XY);
    index2 = ran3l(&ran_seed2)%NN2 + 1;
    coord2 = MOD(coord1 + disp[index2], XY);

    if(element[coord1] != element[coord2]) {
      type1 = element[coord1].get_type();
      type2 = element[coord2].get_type();

      if(type1 == type2) {
	cout << "droplets are touching each other!\t";
	return true;
      }
      // delta_E due to interaction
      for(int i=1; i<=NNn; i++) {
	coordj = MOD(coord1 + disp[i], XY);
	if(element[coordj] != element[coord1])
	  dN -= 2;
      }
      dEJ = dN*J_ij;

      // delta_E due to area constraint.
      // droplet1 - 1 pixel, droplet2 + 1 pixel.
      // There is no area constraint on medium.
      // This is ensured by the value of "type".
      dEA = (Ja*type1*(2*(element[coord1].get_target_area() -
			  element[coord1].get_area()) + 1) +
	     Ja*type2*(-2*(element[coord2].get_target_area() -
			   element[coord2].get_area()) + 1));

      // flip spins
      if(dEJ<=0 && dEA<=0){
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if((dEJ<=0 && dEA>0) && ran3f(&ran_seed3) < exp(-dEA/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if((dEJ>0 && dEA<=0) && ran3f(&ran_seed3) < exp(-dEJ/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
      else if(ran3f(&ran_seed3) < exp(-(dEJ + dEA)/kT)) {
	if(flip_routine(element[coord1], element[coord2], coord1, (int)dEJ))
	  return true;
      }
    }
  }
  return false;
}	
#endif
